#include "framebuffer.h"

#include <iostream>

#include <math.h>

#include "scene.h"

using namespace std;

FrameBuffer::FrameBuffer(int u0, int v0, int _w, int _h) 
: Fl_Gl_Window(u0, v0, _w, _h, 0), w(_w), h(_h) {

  pix = new unsigned int[w*h];
  zb = new float[w*h];

}

void FrameBuffer::Set(unsigned int color) {

  for (int i = 0; i < w*h; i++) {
    pix[i] = color;
  }

}

void FrameBuffer::SetDepth(float depth0) {

  for (int i = 0; i < w*h; i++) {
    zb[i] = depth0;
  }

}

void FrameBuffer::SetGuarded(int u, int v, unsigned int color) {

  if (u < 0 || u > w-1 || v < 0 || v > h-1)
    return;
  Set(u, v, color);

}

void FrameBuffer::Set(int u, int v, unsigned int color) {

  pix[(h-1-v)*w+u] = color;

}

void FrameBuffer::draw() {

  glDrawPixels(w, h, GL_RGBA, GL_UNSIGNED_BYTE, pix);

}

void FrameBuffer::DrawDiagonalCross(
  int u, int v, unsigned int color, int sizePix) {

  for (int i = 0; i < sizePix; i++) {
    SetGuarded(u-sizePix/2+i, v-sizePix/2+i, color);
    SetGuarded(u+sizePix/2-i, v-sizePix/2+i, color);
  }

}

void FrameBuffer::SetGrid(int checkerSize, unsigned int colors[2]) {

  for (int v = 0; v < h; v++) {
    for (int u = 0; u < w; u++) {
      Set(u, v, colors[(u/checkerSize+v/checkerSize)%2]);
    }
  }

}


int FrameBuffer::handle(int event)  {  

  switch(event) 
  {   
  case FL_KEYBOARD: {
    KeyboardHandle();
    return 0;
  }
  default:
    break;
  }
  return 0;

}

void FrameBuffer::KeyboardHandle() {

  int key = Fl::event_key();      
  switch (key) {
    case FL_Up: {
      scene->TranslateForward();
      break;
    }
    case FL_Down: {
      scene->TranslateBackward();
      break;
    }
    case ',': {
      scene->TranslateLeft();
      break;
    }
    case '.': {
      scene->TranslateRight();
      break;
    }
    case 'z': {
      scene->ZoomIn();
      break;
    }
    default:
      break;
  }

}

void FrameBuffer::Draw3DSegment(V3 P0, V3 col0, V3 P1, V3 col1, PPC *ppc) {

  V3 projPs[2];
  if (!ppc->Project(P0, projPs[0]))
    return;
  if (!ppc->Project(P1, projPs[1]))
    return;

  // screen space color interpolation
  Draw2DSegment(projPs[0], col0, projPs[1], col1);

}

void FrameBuffer::Draw2DSegment(V3 P0, V3 col0, V3 P1, V3 col1) {

  int stepsN;
  if (fabsf(P0[0]-P1[0]) < fabsf(P0[1]-P1[1])) {
    stepsN = (int) ceilf(fabsf(P0[1]-P1[1]));
  }
  else {
    stepsN = (int) ceilf(fabsf(P0[0]-P1[0]));
  }

  stepsN = ((stepsN == 0) ? 1 : stepsN);

  V3 dP = (P1-P0)/(float)(stepsN);
  V3 dcol = (col1-col0)/(float)(stepsN);
  for (int i = 0; i <= stepsN; i++) {
    V3 P = P0 + dP*(float)i;
    V3 col = col0 + dcol*(float)i;
    SetIfCloser(P, col);
  }

}

void FrameBuffer::SetIfCloser(V3 P, V3 col) {

  int u = (int)P[0];
  int v = (int)P[1];

  if (u < 0 || u > w-1 || v < 0 || v > h-1)
    return;

  int uv = (h-1-v)*w+u;
  if (zb[uv] > P[2])
    return;
  zb[uv] = P[2];
  pix[uv] = col.UIntColor();

}
