#include "phc.h"
#include "framebuffer.h"
#include <math.h>

PPC::PPC(float hfov, int _w, int _h, float _nearz, float _farz) : w(_w), h(_h),
  nearz(_nearz), farz(_farz) {

  C = V3(0.0f, 0.0f, 0.0f);
  a = V3(1.0f, 0.0f, 0.0f);
  b = V3(0.0f, -1.0f, 0.0f);
  float tan2 = tanf(hfov/2.0f/180.0f*3.14159f);
  c = V3(-(float)w/2.0f, (float)h/2.0f, -(float)w/2.0f/tan2);

  UpdateInvCam();

}

void PPC::UpdateInvCam() {

  invcam.SetColumn(0, a);
  invcam.SetColumn(1, b);
  invcam.SetColumn(2, c);
  invcam = invcam.Invert();

}

bool PPC::Project(V3 P, V3& projP) {

  V3 q = invcam*(P-C);
  if (q[2] <= 0.0f)
    return false;

  projP[0] = q[0] / q[2];
  projP[1] = q[1] / q[2];
  projP[2] = 1.0f / q[2];
  return true;

}

V3 PPC::GetPoint(float uf, float vf) {

  return C+c+a*uf+b*vf;

}

V3 PPC::GetVD() {

  return (a^b).UnitVector();

}


void PPC::Translate(V3 trans) {

  C = C + trans;

}

void PPC::Zoom(float frel) {

  float zstep = Getf()*(frel-1.0f);
  c = c + GetVD()*zstep;
  UpdateInvCam();

}

float PPC::Getf() {

  return c*GetVD();

}

void PPC::Visualize(PPC *dv, FrameBuffer *fb) {

  V3 N[4];
  N[0] = GetPoint(0.0f, 0.0f, nearz);
  N[1] = GetPoint(0.0f, (float)h, nearz);
  N[2] = GetPoint((float)w, (float)h, nearz);
  N[3] = GetPoint((float)w, 0.0f, nearz);
  V3 F[4];
  F[0] = GetPoint(0.0f, 0.0f, farz);
  F[1] = GetPoint(0.0f, (float)h, farz);
  F[2] = GetPoint((float)w, (float)h, farz);
  F[3] = GetPoint((float)w, 0.0f, farz);

  V3 nearColor(1.0f, 0.0f, 0.0f);
  V3 farColor(0.0f, 1.0f, 0.0f);
  for (int i = 0; i < 4; i++) {
    int _i = (i+1)%4;
    fb->Draw3DSegment(N[i], nearColor, N[_i], nearColor, dv);
    fb->Draw3DSegment(F[i], farColor, F[_i], farColor, dv);
    fb->Draw3DSegment(N[i], nearColor, F[i], farColor, dv);
  }

}


V3 PPC::GetPoint(float uf, float vf, float z) {

  float w = z/Getf();
  return C+(c+a*uf+b*vf)*w;

}

void PPC::PositionAndOrient(V3 newC, V3 newVD, V3 quasiUp) {

  newVD = newVD.UnitVector();
  quasiUp = quasiUp.UnitVector();

  float f = Getf();
  V3 projC;
  Project(C+GetVD()*f, projC);

  V3 newa = (newVD^quasiUp).UnitVector();
  newa = newa*a.Length();
  V3 newb = (newVD^newa).UnitVector();
  newb = newb * b.Length();
  V3 newc = newVD*f-newa*projC[0]-newb*projC[1];

  b = newb;
  a = newa;
  c = newc;
  C = newC;

  UpdateInvCam();

}
